@@ -1,3 +1,5 @@
+%% -*- erlang-indent-level: 4;indent-tabs-mode: nil -*-
+%% ex: ts=4 sw=4 et
 -module(ct1_rt).
 
 -compile(export_all).
@@ -7,10 +9,12 @@ files() ->
     [{create, "ebin/a1.app", app(a1)},
      {copy, "../../rebar", "rebar"},
      {copy, "rebar.config", "rebar.config"},
+     {copy, "app.config", "app.config"},
      {copy, "test_SUITE.erl", "itest/test_SUITE.erl"}].
 
 run(_Dir) ->
     {ok, _} = retest:sh("./rebar compile ct"),
+    {ok, _} = retest:sh("./rebar compile ct -v"),
     ok.
 
 
