@@ -1,3 +1,5 @@
+%% -*- erlang-indent-level: 4;indent-tabs-mode: nil -*-
+%% ex: ts=4 sw=4 et
 -module(tplugins_rt).
 -compile(export_all).
 
@@ -18,7 +20,7 @@ files() ->
      {create, "ebin/fish.app", app(fish, [fish])}
     ].
 
-run(Dir) ->
+run(_Dir) ->
     ?assertMatch({ok, _}, retest_sh:run("./rebar fwibble -v", [])),
     ?assertEqual(false, filelib:is_regular("fwibble.test")),
     Ref = retest:sh("./rebar -C bad.config -v clean", [{async, true}]),
