@@ -27,8 +27,11 @@
 -module(rebar_port_compiler).
 
 -export([compile/2,
-         clean/2,
-         setup_env/1]).
+         clean/2]).
+
+%% for internal use only
+-export([setup_env/1,
+         info/2]).
 
 -include("rebar.hrl").
 
@@ -86,37 +89,22 @@
 %%                           "$CFLAGS -X86Options"}]}
 %%
 
-%% TODO: reconsider keeping both sources and objects once
-%%       support for deprecated options has been remove.
-%%       remove [] as valid value for sources, objects, and opts
-%%       when removing deprecated options.
 -record(spec, {type::'drv' | 'exe',
                target::file:filename(),
-               sources = [] :: [file:filename(), ...] | [],
-               objects = [] :: [file:filename(), ...] | [],
+               sources = [] :: [file:filename(), ...],
+               objects = [] :: [file:filename(), ...],
                opts = [] ::list() | []}).
 
 compile(Config, AppFile) ->
-    rebar_utils:deprecated(port_sources, port_specs, Config, "soon"),
-    rebar_utils:deprecated(so_name, port_specs, Config, "soon"),
-    rebar_utils:deprecated(so_specs, port_specs, Config, "soon"),
-
-    %% TODO: remove SpecType and OldSources make get_specs/2
-    %%       return list(#spec{}) when removing deprecated options
-    {SpecType, {OldSources, Specs}} = get_specs(Config, AppFile),
-
-    case {SpecType, OldSources, Specs} of
-        {old, [], _} ->
-            ok;                                 % old specs empty
-        {new, [], []} ->
-            ok;                                 % port_specs empty
-
-        _ ->                                    % have old/new specs
-
-            SharedEnv = rebar_config:get_env(Config, ?MODULE),
+    case get_specs(Config, AppFile) of
+        [] ->
+            ok;
+        Specs ->
+            SharedEnv = rebar_config:get_env(Config, rebar_deps) ++
+                rebar_config:get_env(Config, ?MODULE),
 
             %% Compile each of the sources
-            NewBins = compile_sources(OldSources, Specs, SharedEnv),
+            NewBins = compile_sources(Config, Specs, SharedEnv),
 
             %% Make sure that the target directories exist
             ?INFO("Using specs ~p\n", [Specs]),
@@ -147,23 +135,17 @@ compile(Config, AppFile) ->
     end.
 
 clean(Config, AppFile) ->
-    %% TODO: remove SpecType and OldSources make get_specs/2
-    %%       return list(#spec{}) when removing deprecated options
-    {SpecType, {OldSources, Specs}} = get_specs(Config, AppFile),
-
-    case {SpecType, OldSources, Specs} of
-        {old, [], _} ->
-            ok;                                 % old specs empty
-        {new, [], []} ->
-            ok;                                 % port_specs empty
-
-        _ ->                                    % have old/new specs
-
+    case get_specs(Config, AppFile) of
+        [] ->
+            ok;
+        Specs ->
             lists:foreach(fun(#spec{target=Target, objects=Objects}) ->
                                   rebar_file_utils:delete_each([Target]),
-                                  rebar_file_utils:delete_each(Objects)
+                                  rebar_file_utils:delete_each(Objects),
+                                  rebar_file_utils:delete_each(port_deps(Objects))
                           end, Specs)
-    end.
+    end,
+    ok.
 
 setup_env(Config) ->
     setup_env(Config, []).
@@ -172,20 +154,48 @@ setup_env(Config) ->
 %% Internal functions
 %% ===================================================================
 
+info(help, compile) ->
+    info_help("Build port sources");
+info(help, clean) ->
+    info_help("Delete port build results").
+
+info_help(Description) ->
+    ?CONSOLE(
+       "~s.~n"
+       "~n"
+       "Valid rebar.config options:~n"
+       "  ~p~n"
+       "  ~p~n",
+       [
+        Description,
+        {port_env, [{"CFLAGS", "$CFLAGS -Ifoo"},
+                    {"freebsd", "LDFLAGS", "$LDFLAGS -lfoo"}]},
+        {port_specs, [{"priv/so_name.so", ["c_src/*.c"]},
+                      {"linux", "priv/hello_linux", ["c_src/hello_linux.c"]},
+                      {"linux", "priv/hello_linux", ["c_src/*.c"], [{env, []}]}]}
+       ]).
+
 setup_env(Config, ExtraEnv) ->
     %% Extract environment values from the config (if specified) and
     %% merge with the default for this operating system. This enables
     %% max flexibility for users.
     DefaultEnv  = filter_env(default_env(), []),
-    PortEnv = filter_env(port_env(Config), []),
-    OverrideEnv = global_defines() ++ PortEnv ++ filter_env(ExtraEnv, []),
+
+    %% Get any port-specific envs; use port_env first and then fallback
+    %% to port_envs for compatibility
+    RawPortEnv = rebar_config:get_list(Config, port_env,
+                          rebar_config:get_list(Config, port_envs, [])),
+
+    PortEnv = filter_env(RawPortEnv, []),
+    Defines = get_defines(Config),
+    OverrideEnv = Defines ++ PortEnv ++ filter_env(ExtraEnv, []),
     RawEnv = apply_defaults(os_env(), DefaultEnv) ++ OverrideEnv,
     expand_vars_loop(merge_each_var(RawEnv, [])).
 
-global_defines() ->
-    Defines = rebar_config:get_global(defines, []),
-    Flags = string:join(["-D" ++ D || D <- Defines], " "),
-    [{"ERL_CFLAGS", "$ERL_CFLAGS " ++ Flags}].
+get_defines(Config) ->
+    RawDefines = rebar_config:get_xconf(Config, defines, []),
+    Defines = string:join(["-D" ++ D || D <- RawDefines], " "),
+    [{"ERL_CFLAGS", "$ERL_CFLAGS " ++ Defines}].
 
 replace_extension(File, NewExt) ->
     OldExt = filename:extension(File),
@@ -198,36 +208,72 @@ replace_extension(File, OldExt, NewExt) ->
 %% == compile and link ==
 %%
 
-compile_sources([], Specs, SharedEnv) ->        % port_spec
+compile_sources(Config, Specs, SharedEnv) ->
     lists:foldl(
       fun(#spec{sources=Sources, type=Type, opts=Opts}, NewBins) ->
               Env = proplists:get_value(env, Opts, SharedEnv),
-              compile_each(Sources, Type, Env, NewBins)
-      end, [], Specs);
-compile_sources(OldSources, _Specs, SharedEnv) -> % deprecated
-    compile_each(OldSources, drv, SharedEnv, []).
+              compile_each(Config, Sources, Type, Env, NewBins)
+      end, [], Specs).
 
-compile_each([], _Type, _Env, NewBins) ->
+compile_each(_Config, [], _Type, _Env, NewBins) ->
     lists:reverse(NewBins);
-compile_each([Source | Rest], Type, Env, NewBins) ->
+compile_each(Config, [Source | Rest], Type, Env, NewBins) ->
     Ext = filename:extension(Source),
     Bin = replace_extension(Source, Ext, ".o"),
     case needs_compile(Source, Bin) of
         true ->
-            ?CONSOLE("Compiling ~s\n", [Source]),
             Template = select_compile_template(Type, compiler(Ext)),
-            rebar_utils:sh(expand_command(Template, Env, Source, Bin),
-                           [{env, Env}]),
-            compile_each(Rest, Type, Env, [Bin | NewBins]);
+            Cmd = expand_command(Template, Env, Source, Bin),
+            ShOpts = [{env, Env}, return_on_error, {use_stdout, false}],
+            exec_compiler(Config, Source, Cmd, ShOpts),
+            compile_each(Config, Rest, Type, Env, [Bin | NewBins]);
         false ->
             ?INFO("Skipping ~s\n", [Source]),
-            compile_each(Rest, Type, Env, NewBins)
+            compile_each(Config, Rest, Type, Env, NewBins)
+    end.
+
+exec_compiler(Config, Source, Cmd, ShOpts) ->
+    case rebar_utils:sh(Cmd, ShOpts) of
+        {error, {_RC, RawError}} ->
+            AbsSource = case rebar_utils:processing_base_dir(Config) of
+                            true ->
+                                Source;
+                            false ->
+                                filename:absname(Source)
+                        end,
+            ?CONSOLE("Compiling ~s\n", [AbsSource]),
+            Error = re:replace(RawError, Source, AbsSource,
+                               [{return, list}, global]),
+            ?CONSOLE("~s", [Error]),
+            ?FAIL;
+        {ok, Output} ->
+            ?CONSOLE("Compiling ~s\n", [Source]),
+            ?CONSOLE("~s", [Output])
     end.
 
 needs_compile(Source, Bin) ->
-    %% TODO: Generate depends using gcc -MM so we can also
-    %% check for include changes
-    filelib:last_modified(Bin) < filelib:last_modified(Source).
+    needs_link(Bin, [Source|bin_deps(Bin)]).
+
+%% NOTE: This relies on -MMD being passed to the compiler and returns an
+%% empty list if the .d file is not available.  This means header deps are
+%% ignored on win32.
+bin_deps(Bin) ->
+    [DepFile] = port_deps([Bin]),
+    case file:read_file(DepFile) of
+        {ok, Deps} ->
+            Ds = parse_bin_deps(list_to_binary(Bin), Deps),
+            ?DEBUG("Deps of ~p: ~p\n", [Bin, Ds]),
+            Ds;
+        {error, Err} ->
+            ?DEBUG("Skipping deps parse of ~s: ~p\n", [DepFile, Err]),
+            []
+    end.
+
+parse_bin_deps(Bin, Deps) ->
+    Sz = size(Bin),
+    <<Bin:Sz/binary, ": ", X/binary>> = Deps,
+    Ds = re:split(X, "\\s*\\\\\\R\\s*|\\s+", [{return, binary}]),
+    [D || D <- Ds, D =/= <<>>].
 
 needs_link(SoName, []) ->
     filelib:last_modified(SoName) == 0;
@@ -247,18 +293,37 @@ needs_link(SoName, NewBins) ->
 %%
 
 get_specs(Config, AppFile) ->
-    case rebar_config:get_local(Config, port_specs, undefined) of
-        undefined ->
-            %% TODO: DEPRECATED: remove support for non-port_specs syntax
-            {old, old_get_specs(Config, AppFile)};
-        PortSpecs ->
-            {new, get_port_specs(Config, PortSpecs)}
-    end.
+    Specs = case rebar_config:get_local(Config, port_specs, []) of
+                [] ->
+                    %% No spec provided. Construct a spec
+                    %% from old-school so_name and sources
+                    [port_spec_from_legacy(Config, AppFile)];
+                PortSpecs ->
+                    Filtered = filter_port_specs(PortSpecs),
+                    OsType = os:type(),
+                    [get_port_spec(Config, OsType, Spec) || Spec <- Filtered]
+            end,
+    [S || S <- Specs, S#spec.sources /= []].
 
-get_port_specs(Config, PortSpecs) ->
-    Filtered = filter_port_specs(PortSpecs),
-    OsType = os:type(),
-    {[], [get_port_spec(Config, OsType, Spec) || Spec <- Filtered]}.
+port_spec_from_legacy(Config, AppFile) ->
+    %% Get the target from the so_name variable
+    Target = case rebar_config:get(Config, so_name, undefined) of
+                 undefined ->
+                     %% Generate a sensible default from app file
+                     {_, AppName} = rebar_app_utils:app_name(Config, AppFile),
+                     filename:join("priv",
+                                   lists:concat([AppName, "_drv.so"]));
+                 AName ->
+                     %% Old form is available -- use it
+                     filename:join("priv", AName)
+             end,
+    %% Get the list of source files from port_sources
+    Sources = port_sources(rebar_config:get_list(Config, port_sources,
+                                                 ["c_src/*.c"])),
+    #spec { type = target_type(Target),
+            target = maybe_switch_extension(os:type(), Target),
+            sources = Sources,
+            objects = port_objects(Sources) }.
 
 filter_port_specs(Specs) ->
     [S || S <- Specs, filter_port_spec(S)].
@@ -289,6 +354,9 @@ port_sources(Sources) ->
 port_objects(SourceFiles) ->
     [replace_extension(O, ".o") || O <- SourceFiles].
 
+port_deps(SourceFiles) ->
+    [replace_extension(O, ".d") || O <- SourceFiles].
+
 port_opts(Config, Opts) ->
     [port_opt(Config, O) || O <- Opts].
 
@@ -309,58 +377,6 @@ switch_to_dll_or_exe(Target) ->
         _Other -> Target
     end.
 
-%% TODO: DEPRECATED: remove support for non-port_specs syntax [old_*()]
-old_get_specs(Config, AppFile) ->
-    OsType = os:type(),
-    SourceFiles = old_get_sources(Config),
-    Specs =
-        case rebar_config:get_local(Config, so_specs, undefined) of
-            undefined ->
-                Objects = port_objects(SourceFiles),
-                %% New form of so_specs is not provided. See if the old form
-                %% of {so_name} is available instead
-                Dir = "priv",
-                SoName =
-                    case rebar_config:get_local(Config, so_name, undefined) of
-                        undefined ->
-                            %% Ok, neither old nor new form is
-                            %% available. Use the app name and
-                            %% generate a sensible default.
-                            AppName = rebar_app_utils:app_name(AppFile),
-                            DrvName = ?FMT("~s_drv.so", [AppName]),
-                            filename:join([Dir, DrvName]);
-                        AName ->
-                            %% Old form is available -- use it
-                            filename:join(Dir, AName)
-                    end,
-                [old_get_so_spec({SoName, Objects}, OsType)];
-            SoSpecs ->
-                [old_get_so_spec(S, OsType) || S <- SoSpecs]
-        end,
-    {SourceFiles, Specs}.
-
-old_get_sources(Config) ->
-    RawSources = rebar_config:get_local(Config, port_sources,
-                                        ["c_src/*.c"]),
-    FilteredSources = old_filter_port_sources(RawSources),
-    old_expand_sources(FilteredSources).
-
-old_filter_port_sources(PortSources) ->
-    [S || S <- PortSources, old_is_arch_port_sources(S)].
-
-old_is_arch_port_sources({Arch, _Sources}) -> rebar_utils:is_arch(Arch);
-old_is_arch_port_sources(_Sources)         -> true.
-
-old_expand_sources(Sources) ->
-    lists:flatmap(fun filelib:wildcard/1, Sources).
-
-old_get_so_spec({Target, Objects}, OsType) ->
-    #spec{type=drv,
-          target=maybe_switch_extension(OsType, Target),
-          sources=[],
-          objects=Objects,
-          opts=[]}.
-
 %%
 %% == port_env ==
 %%
@@ -431,7 +447,7 @@ expand_vars_loop([], Recurse, Vars, Count) ->
     expand_vars_loop(Recurse, [], Vars, Count-1);
 expand_vars_loop([{K, V} | Rest], Recurse, Vars, Count) ->
     %% Identify the variables that need expansion in this value
-    ReOpts = [global, {capture, all_but_first, list}],
+    ReOpts = [global, {capture, all_but_first, list}, unicode],
     case re:run(V, "\\\${?(\\w+)}?", ReOpts) of
         {match, Matches} ->
             %% Identify the unique variables that need to be expanded
@@ -473,8 +489,7 @@ expand_keys_in_value([Key | Rest], Value, Vars) ->
 expand_command(TmplName, Env, InFiles, OutFile) ->
     Cmd0 = proplists:get_value(TmplName, Env),
     Cmd1 = rebar_utils:expand_env_variable(Cmd0, "PORT_IN_FILES", InFiles),
-    Cmd2 = rebar_utils:expand_env_variable(Cmd1, "PORT_OUT_FILE", OutFile),
-    re:replace(Cmd2, "\\\$\\w+|\\\${\\w+}", "", [global, {return, list}]).
+    rebar_utils:expand_env_variable(Cmd1, "PORT_OUT_FILE", OutFile).
 
 %%
 %% Given a string, determine if it is expandable
@@ -485,35 +500,6 @@ is_expandable(InStr) ->
         nomatch -> false
     end.
 
-port_env(Config) ->
-    %% TODO: remove support for deprecated port_envs option
-    PortEnv = rebar_utils:get_deprecated_list(Config, port_envs, port_env,
-                                              [], "soon"),
-    %% TODO: remove migration of deprecated port_env DRV_-/EXE_-less vars
-    %%       when the deprecation grace period ends
-    WarnAndConvertVar = fun(Var) ->
-                                New = "DRV_" ++ Var,
-                                rebar_utils:deprecated(Var, New, "soon"),
-                                New
-                        end,
-    ConvertVar = fun(Var="CXX_TEMPLATE") -> WarnAndConvertVar(Var);
-                    (Var="CC_TEMPLATE")  -> WarnAndConvertVar(Var);
-                    (Var="LINK_TEMPLATE") -> WarnAndConvertVar(Var);
-                    (Var) -> Var
-                 end,
-    %% Also warn about references to deprecated vars? omitted for
-    %% performance reasons.
-    ReplaceVars = fun(Val) ->
-                          re:replace(Val, "\\$(CXX|CC|LINK)(_TEMPLATE)",
-                                     "DRV_\\1\\2", [{return,list}, global])
-                  end,
-    Convert = fun({ArchRegex, Var, Val}) ->
-                      {ArchRegex, ConvertVar(Var), ReplaceVars(Val)};
-                 ({Var, Val}) ->
-                      {ConvertVar(Var), ReplaceVars(Val)}
-              end,
-    [Convert(EnvVar) || EnvVar <- PortEnv].
-
 %%
 %% Filter a list of env vars such that only those which match the provided
 %% architecture regex (or do not have a regex) are returned.
@@ -534,11 +520,21 @@ erts_dir() ->
     lists:concat([code:root_dir(), "/erts-", erlang:system_info(version)]).
 
 os_env() ->
-    Os = [list_to_tuple(re:split(S, "=", [{return, list}, {parts, 2}])) ||
-             S <- os:getenv()],
+    ReOpts = [{return, list}, {parts, 2}, unicode],
+    Os = [list_to_tuple(re:split(S, "=", ReOpts)) ||
+             S <- lists:filter(fun discard_deps_vars/1, os:getenv())],
     %% Drop variables without a name (win32)
     [T1 || {K, _V} = T1 <- Os, K =/= []].
 
+%%
+%% To avoid having multiple repetitions of the same environment variables
+%% (ERL_LIBS), avoid exporting any variables that may cause conflict with
+%% those exported by the rebar_deps module (ERL_LIBS, REBAR_DEPS_DIR)
+%%
+discard_deps_vars("ERL_LIBS=" ++ _Value)       -> false;
+discard_deps_vars("REBAR_DEPS_DIR=" ++ _Value) -> false;
+discard_deps_vars(_Var)                        -> true.
+
 select_compile_template(drv, Compiler) ->
     select_compile_drv_template(Compiler);
 select_compile_template(exe, Compiler) ->
@@ -587,15 +583,15 @@ default_env() ->
       "$CC -c $CFLAGS $EXE_CFLAGS $PORT_IN_FILES -o $PORT_OUT_FILE"},
      {"EXE_LINK_TEMPLATE",
       "$CC $PORT_IN_FILES $LDFLAGS $EXE_LDFLAGS -o $PORT_OUT_FILE"},
-     {"DRV_CFLAGS" , "-g -Wall -fPIC $ERL_CFLAGS"},
+     {"DRV_CFLAGS" , "-g -Wall -fPIC -MMD $ERL_CFLAGS"},
      {"DRV_LDFLAGS", "-shared $ERL_LDFLAGS"},
-     {"EXE_CFLAGS" , "-g -Wall -fPIC $ERL_CFLAGS"},
+     {"EXE_CFLAGS" , "-g -Wall -fPIC -MMD $ERL_CFLAGS"},
      {"EXE_LDFLAGS", "$ERL_LDFLAGS"},
 
-     {"ERL_CFLAGS", lists:concat([" -I", erl_interface_dir(include),
-                                  " -I", filename:join(erts_dir(), "include"),
-                                  " "])},
-     {"ERL_EI_LIBDIR", erl_interface_dir(lib)},
+     {"ERL_CFLAGS", lists:concat([" -I\"", erl_interface_dir(include),
+                                  "\" -I\"", filename:join(erts_dir(), "include"),
+                                  "\" "])},
+     {"ERL_EI_LIBDIR", lists:concat(["\"", erl_interface_dir(lib), "\""])},
      {"ERL_LDFLAGS"  , " -L$ERL_EI_LIBDIR -lerl_interface -lei"},
      {"ERLANG_ARCH"  , rebar_utils:wordsize()},
      {"ERLANG_TARGET", rebar_utils:get_arch()},
@@ -608,18 +604,42 @@ default_env() ->
      {"solaris.*-64$", "CXXFLAGS", "-D_REENTRANT -m64 $CXXFLAGS"},
      {"solaris.*-64$", "LDFLAGS", "-m64 $LDFLAGS"},
 
+     %% Linux specific flags for multiarch
+     {"linux.*-64$", "CFLAGS", "-m64 $CFLAGS"},
+     {"linux.*-64$", "CXXFLAGS", "-m64 $CXXFLAGS"},
+     {"linux.*-64$", "LDFLAGS", "$LDFLAGS"},
+
      %% OS X Leopard flags for 64-bit
      {"darwin9.*-64$", "CFLAGS", "-m64 $CFLAGS"},
      {"darwin9.*-64$", "CXXFLAGS", "-m64 $CXXFLAGS"},
      {"darwin9.*-64$", "LDFLAGS", "-arch x86_64 $LDFLAGS"},
 
-     %% OS X Snow Leopard flags for 32-bit
-     {"darwin10.*-32", "CFLAGS", "-m32 $CFLAGS"},
-     {"darwin10.*-32", "CXXFLAGS", "-m32 $CXXFLAGS"},
-     {"darwin10.*-32", "LDFLAGS", "-arch i386 $LDFLAGS"},
+     %% OS X Snow Leopard, Lion, and Mountain Lion flags for 32-bit
+     {"darwin1[0-2].*-32", "CFLAGS", "-m32 $CFLAGS"},
+     {"darwin1[0-2].*-32", "CXXFLAGS", "-m32 $CXXFLAGS"},
+     {"darwin1[0-2].*-32", "LDFLAGS", "-arch i386 $LDFLAGS"},
 
-     %% OS X Lion flags for 32-bit
-     {"darwin11.*-32", "CFLAGS", "-m32 $CFLAGS"},
-     {"darwin11.*-32", "CXXFLAGS", "-m32 $CXXFLAGS"},
-     {"darwin11.*-32", "LDFLAGS", "-arch i386 $LDFLAGS"}
+     %% Windows specific flags
+     %% add MS Visual C++ support to rebar on Windows
+     {"win32", "CC", "cl.exe"},
+     {"win32", "CXX", "cl.exe"},
+     {"win32", "LINKER", "link.exe"},
+     {"win32", "DRV_CXX_TEMPLATE",
+      %% DRV_* and EXE_* Templates are identical
+      "$CXX /c $CXXFLAGS $DRV_CFLAGS $PORT_IN_FILES /Fo$PORT_OUT_FILE"},
+     {"win32", "DRV_CC_TEMPLATE",
+      "$CC /c $CFLAGS $DRV_CFLAGS $PORT_IN_FILES /Fo$PORT_OUT_FILE"},
+     {"win32", "DRV_LINK_TEMPLATE",
+      "$LINKER $PORT_IN_FILES $LDFLAGS $DRV_LDFLAGS /OUT:$PORT_OUT_FILE"},
+     %% DRV_* and EXE_* Templates are identical
+     {"win32", "EXE_CXX_TEMPLATE",
+      "$CXX /c $CXXFLAGS $EXE_CFLAGS $PORT_IN_FILES /Fo$PORT_OUT_FILE"},
+     {"win32", "EXE_CC_TEMPLATE",
+      "$CC /c $CFLAGS $EXE_CFLAGS $PORT_IN_FILES /Fo$PORT_OUT_FILE"},
+     {"win32", "EXE_LINK_TEMPLATE",
+      "$LINKER $PORT_IN_FILES $LDFLAGS $EXE_LDFLAGS /OUT:$PORT_OUT_FILE"},
+     %% ERL_CFLAGS are ok as -I even though strictly it should be /I
+     {"win32", "ERL_LDFLAGS", " /LIBPATH:$ERL_EI_LIBDIR erl_interface.lib ei.lib"},
+     {"win32", "DRV_CFLAGS", "/Zi /Wall $ERL_CFLAGS"},
+     {"win32", "DRV_LDFLAGS", "/DLL $ERL_LDFLAGS"}
     ].
