@@ -31,24 +31,48 @@
          get_arch/0,
          wordsize/0,
          sh/2,
+         sh_send/3,
          find_files/2,
          find_files/3,
+         find_files_by_ext/2,
+         find_files_by_ext/3,
          now_str/0,
          ensure_dir/1,
-         beam_to_mod/2, beams/1,
+         beam_to_mod/2,
+         beams/1,
          erl_to_mod/1,
+         abort/0,
          abort/2,
          escript_foldl/3,
          find_executable/1,
          prop_check/3,
          expand_code_path/0,
-         deprecated/3, deprecated/4,
          expand_env_variable/3,
-         vcs_vsn/2,
-         get_deprecated_global/3,
-         get_deprecated_list/4, get_deprecated_list/5,
-         get_deprecated_local/4, get_deprecated_local/5,
-         delayed_halt/1]).
+         vcs_vsn/3,
+         deprecated/3,
+         deprecated/4,
+         get_deprecated_global/4,
+         get_deprecated_global/5,
+         get_experimental_global/3,
+         get_experimental_local/3,
+         get_deprecated_list/4,
+         get_deprecated_list/5,
+         get_deprecated_local/4,
+         get_deprecated_local/5,
+         delayed_halt/1,
+         erl_opts/1,
+         src_dirs/1,
+         ebin_dir/0,
+         base_dir/1,
+         processing_base_dir/1,
+         processing_base_dir/2,
+         patch_env/2,
+         cleanup_code_path/1]).
+
+%% for internal use only
+-export([otp_release/0]).
+
+-dialyzer({no_missing_calls, escript_foldl/3}).
 
 -include("rebar.hrl").
 
@@ -70,7 +94,7 @@ is_arch(ArchRegex) ->
 
 get_arch() ->
     Words = wordsize(),
-    erlang:system_info(otp_release) ++ "-"
+    otp_release() ++ "-"
         ++ erlang:system_info(system_architecture) ++ "-" ++ Words.
 
 wordsize() ->
@@ -82,6 +106,25 @@ wordsize() ->
             integer_to_list(8 * erlang:system_info(wordsize))
     end.
 
+sh_send(Command0, String, Options0) ->
+    ?INFO("sh_send info:\n\tcwd: ~p\n\tcmd: ~s < ~s\n",
+          [get_cwd(), Command0, String]),
+    ?DEBUG("\topts: ~p\n", [Options0]),
+
+    DefaultOptions = [use_stdout, abort_on_error],
+    Options = [expand_sh_flag(V)
+               || V <- proplists:compact(Options0 ++ DefaultOptions)],
+
+    Command = patch_on_windows(Command0, proplists:get_value(env, Options, [])),
+    PortSettings = proplists:get_all_values(port_settings, Options) ++
+        [exit_status, {line, 16384}, use_stdio, stderr_to_stdout, hide],
+    Port = open_port({spawn, Command}, PortSettings),
+
+    %% allow us to send some data to the shell command's STDIN
+    %% Erlang doesn't let us get any reply after sending an EOF, though...
+    Port ! {self(), {command, String}},
+    port_close(Port).
+
 %%
 %% Options = [Option] -- defaults to [use_stdout, abort_on_error]
 %% Option = ErrorOption | OutputOption | {cd, string()} | {env, Env}
@@ -104,6 +147,7 @@ sh(Command0, Options0) ->
     Command = patch_on_windows(Command0, proplists:get_value(env, Options, [])),
     PortSettings = proplists:get_all_values(port_settings, Options) ++
         [exit_status, {line, 16384}, use_stdio, stderr_to_stdout, hide],
+    ?DEBUG("Port Cmd: ~p\nPort Opts: ~p\n", [Command, PortSettings]),
     Port = open_port({spawn, Command}, PortSettings),
 
     case sh_loop(Port, OutputHandler, []) of
@@ -120,6 +164,28 @@ find_files(Dir, Regex, Recursive) ->
     filelib:fold_files(Dir, Regex, Recursive,
                        fun(F, Acc) -> [F | Acc] end, []).
 
+%% Find files by extension, for example ".erl", avoiding resource fork
+%% files in OS X.  Such files are named for example src/._xyz.erl
+%% Such files may also appear with network filesystems on OS X.
+%%
+%% The Ext is really a regexp, with any leading dot implicitly
+%% escaped, and anchored at the end of the string.
+%%
+find_files_by_ext(Dir, Ext) ->
+    find_files_by_ext(Dir, Ext, true).
+
+find_files_by_ext(Dir, Ext, Recursive) ->
+    %% Convert simple extension to proper regex
+    EscapeDot = case Ext of
+                    "." ++ _ ->
+                        "\\";
+                    _ ->
+                        %% allow for other suffixes, such as _pb.erl
+                        ""
+                end,
+    ExtRe = "^[^._].*" ++ EscapeDot ++ Ext ++ [$$],
+    find_files(Dir, ExtRe, Recursive).
+
 now_str() ->
     {{Year, Month, Day}, {Hour, Minute, Second}} = calendar:local_time(),
     lists:flatten(io_lib:format("~4b/~2..0b/~2..0b ~2..0b:~2..0b:~2..0b",
@@ -137,10 +203,14 @@ ensure_dir(Path) ->
             Error
     end.
 
+-spec abort() -> no_return().
+abort() ->
+    throw(rebar_abort).
+
 -spec abort(string(), [term()]) -> no_return().
 abort(String, Args) ->
     ?ERROR(String, Args),
-    delayed_halt(1).
+    abort().
 
 %% TODO: Rename emulate_escript_foldl to escript_foldl and remove
 %% this function when the time is right. escript:foldl/3 was an
@@ -167,9 +237,11 @@ prop_check(false, Msg, Args) -> ?ABORT(Msg, Args).
 
 %% Convert all the entries in the code path to absolute paths.
 expand_code_path() ->
-    CodePath = lists:foldl(fun (Path, Acc) ->
-                                   [filename:absname(Path) | Acc]
-                           end, [], code:get_path()),
+    CodePath = lists:foldl(
+                 fun(Path, Acc) ->
+                         Path1 = rmemo:call(filename, absname, [Path]),
+                         [Path1 | Acc]
+                 end, [], code:get_path()),
     code:set_path(lists:reverse(CodePath)).
 
 %%
@@ -183,74 +255,39 @@ expand_env_variable(InStr, VarName, RawVarValue) ->
             %% No variables to expand
             InStr;
         _ ->
-            VarValue = re:replace(RawVarValue, "\\\\", "\\\\\\\\", [global]),
+            ReOpts = [global, unicode, {return, list}],
+            VarValue = re:replace(RawVarValue, "\\\\", "\\\\\\\\", ReOpts),
             %% Use a regex to match/replace:
-            %% Given variable "FOO": match $FOO\s | $FOOeol | ${FOO}
-            RegEx = io_lib:format("\\\$(~s(\\s|$)|{~s})", [VarName, VarName]),
-            ReOpts = [global, {return, list}],
+            %% Given variable "FOO", match $FOO\W | $FOOeol | ${FOO}.
+            RegEx = io_lib:format("\\\$(~s(\\W|$)|{~s})", [VarName, VarName]),
             re:replace(InStr, RegEx, [VarValue, "\\2"], ReOpts)
     end.
 
-vcs_vsn(Vcs, Dir) ->
-    Key = {Vcs, Dir},
-    try ets:lookup_element(rebar_vsn_cache, Key, 2)
-    catch
-        error:badarg ->
-            VsnString = vcs_vsn_1(Vcs, Dir),
-            ets:insert(rebar_vsn_cache, {Key, VsnString}),
-            VsnString
+vcs_vsn(Config, Vsn, Dir) ->
+    Key = {Vsn, Dir},
+    Cache = rebar_config:get_xconf(Config, vsn_cache),
+    case dict:find(Key, Cache) of
+        error ->
+            VsnString = vcs_vsn_1(Vsn, Dir),
+            Cache1 = dict:store(Key, VsnString, Cache),
+            Config1 = rebar_config:set_xconf(Config, vsn_cache, Cache1),
+            {Config1, VsnString};
+        {ok, VsnString} ->
+            {Config, VsnString}
     end.
 
-vcs_vsn_1(Vcs, Dir) ->
-    case vcs_vsn_cmd(Vcs) of
-        {unknown, VsnString} ->
-            ?DEBUG("vcs_vsn: Unknown VCS atom in vsn field: ~p\n", [Vcs]),
-            VsnString;
-        {cmd, CmdString} ->
-            vcs_vsn_invoke(CmdString, Dir);
-        Cmd ->
-            %% If there is a valid VCS directory in the application directory,
-            %% use that version info
-            Extension = lists:concat([".", Vcs]),
-            case filelib:is_dir(filename:join(Dir, Extension)) of
-                true ->
-                    ?DEBUG("vcs_vsn: Primary vcs used for ~s\n", [Dir]),
-                    vcs_vsn_invoke(Cmd, Dir);
-                false ->
-                    %% No VCS directory found for the app. Depending on source
-                    %% tree structure, there may be one higher up, but that can
-                    %% yield unexpected results when used with deps. So, we
-                    %% fallback to searching for a priv/vsn.Vcs file.
-                    VsnFile = filename:join([Dir, "priv", "vsn" ++ Extension]),
-                    case file:read_file(VsnFile) of
-                        {ok, VsnBin} ->
-                            ?DEBUG("vcs_vsn: Read ~s from priv/vsn.~p\n",
-                                   [VsnBin, Vcs]),
-                            string:strip(binary_to_list(VsnBin), right, $\n);
-                        {error, enoent} ->
-                            ?DEBUG("vcs_vsn: Fallback to vcs for ~s\n", [Dir]),
-                            vcs_vsn_invoke(Cmd, Dir)
-                    end
-            end
-    end.
+get_deprecated_global(Config, OldOpt, NewOpt, When) ->
+    get_deprecated_global(Config, OldOpt, NewOpt, undefined, When).
 
-get_deprecated_global(OldOpt, NewOpt, When) ->
-    get_deprecated_global(OldOpt, NewOpt, undefined, When).
+get_deprecated_global(Config, OldOpt, NewOpt, Default, When) ->
+    get_deprecated_3(fun rebar_config:get_global/3,
+                     Config, OldOpt, NewOpt, Default, When).
 
-get_deprecated_global(OldOpt, NewOpt, Default, When) ->
-    case rebar_config:get_global(NewOpt, Default) of
-        undefined ->
-            case rebar_config:get_global(OldOpt, Default) of
-                undefined ->
-                    undefined;
-                Old ->
-                    deprecated(OldOpt, NewOpt, When),
-                    Old
-            end;
-        New ->
-            New
-    end.
+get_experimental_global(Config, Opt, Default) ->
+    get_experimental_3(fun rebar_config:get_global/3, Config, Opt, Default).
 
+get_experimental_local(Config, Opt, Default) ->
+    get_experimental_3(fun rebar_config:get_local/3, Config, Opt, Default).
 
 get_deprecated_list(Config, OldOpt, NewOpt, When) ->
     get_deprecated_list(Config, OldOpt, NewOpt, undefined, When).
@@ -310,10 +347,126 @@ delayed_halt(Code) ->
             end
     end.
 
+%% @doc Return list of erl_opts
+-spec erl_opts(rebar_config:config()) -> list().
+erl_opts(Config) ->
+    RawErlOpts = filter_defines(rebar_config:get(Config, erl_opts, []), []),
+    Defines = [{d, list_to_atom(D)} ||
+                  D <- rebar_config:get_xconf(Config, defines, [])],
+    Opts = Defines ++ RawErlOpts,
+    case proplists:is_defined(no_debug_info, Opts) of
+        true ->
+            [O || O <- Opts, O =/= no_debug_info];
+        false ->
+            [debug_info|Opts]
+    end.
+
+-spec src_dirs([string()]) -> [file:filename(), ...].
+src_dirs([]) ->
+    ["src"];
+src_dirs(SrcDirs) ->
+    SrcDirs.
+
+ebin_dir() ->
+    filename:join(get_cwd(), "ebin").
+
+base_dir(Config) ->
+    rebar_config:get_xconf(Config, base_dir).
+
+processing_base_dir(Config) ->
+    Cwd = rebar_utils:get_cwd(),
+    processing_base_dir(Config, Cwd).
+
+processing_base_dir(Config, Dir) ->
+    AbsDir = filename:absname(Dir),
+    AbsDir =:= base_dir(Config).
+
+%% @doc Returns the list of environment variables including 'REBAR' which
+%% points to the rebar executable used to execute the currently running
+%% command. The environment is not modified if rebar was invoked
+%% programmatically.
+-spec patch_env(rebar_config:config(), [{string(), string()}])
+               -> [{string(), string()}].
+patch_env(Config, []) ->
+    %% If we reached an empty list, the env did not contain the REBAR variable.
+    case rebar_config:get_xconf(Config, escript, "") of
+        "" -> % rebar was invoked programmatically
+            [];
+        Path ->
+            [{"REBAR", Path}]
+    end;
+patch_env(_Config, [{"REBAR", _} | _]=All) ->
+    All;
+patch_env(Config, [E | Rest]) ->
+    [E | patch_env(Config, Rest)].
+
 %% ====================================================================
 %% Internal functions
 %% ====================================================================
 
+otp_release() ->
+    rmemo:call(fun otp_release_1/1, [(erlang:system_info(otp_release))]).
+
+%% If OTP <= R16, otp_release is already what we want.
+otp_release_1([$R,N|_]=Rel) when is_integer(N) ->
+    Rel;
+%% If OTP >= 17.x, erlang:system_info(otp_release) returns just the
+%% major version number, we have to read the full version from
+%% a file. See http://www.erlang.org/doc/system_principles/versions.html
+otp_release_1(Rel) ->
+    Files = [
+             filename:join([code:root_dir(), "releases", Rel, "OTP_VERSION"]),
+             filename:join([code:root_dir(), "OTP_VERSION"])
+            ],
+
+    %% It's possible that none of the above files exist on the filesystem, in
+    %% which case, we're just going to rely on the provided "Rel" (which should
+    %% just be the value of `erlang:system_info(otp_release)`).
+    case read_otp_version_files(Files) of
+        undefined ->
+            warn_missing_otp_version_file(Rel),
+            Rel;
+        Vsn ->
+            Vsn
+    end.
+
+warn_missing_otp_version_file(Rel) ->
+    ?WARN("No OTP_VERSION file found. Using version string ~p.~n", [Rel]).
+
+%% Try to open each file path provided, and if any of them exist on the
+%% filesystem, read their contents and return the value of the first one found.
+read_otp_version_files([]) ->
+    undefined;
+read_otp_version_files([File | Rest]) ->
+    case file:read_file(File) of
+        {ok, Vsn} -> normalize_otp_version(Vsn);
+        {error, enoent} -> read_otp_version_files(Rest)
+    end.
+
+%% Takes the Version binary as read from the OTP_VERSION file and strips any
+%% trailing "**" and trailing "\n", returning the string as a list.
+normalize_otp_version(Vsn) ->
+    %% It's fine to rely on the binary module here because we can
+    %% be sure that it's available when the otp_release string does
+    %% not begin with $R.
+    Size = byte_size(Vsn),
+    %% The shortest vsn string consists of at least two digits
+    %% followed by "\n". Therefore, it's safe to assume Size >= 3.
+    case binary:part(Vsn, {Size, -3}) of
+        <<"**\n">> ->
+            %% The OTP documentation mentions that a system patched
+            %% using the otp_patch_apply tool available to licensed
+            %% customers will leave a '**' suffix in the version as a
+            %% flag saying the system consists of application versions
+            %% from multiple OTP versions. We ignore this flag and
+            %% drop the suffix, given for all intents and purposes, we
+            %% cannot obtain relevant information from it as far as
+            %% tooling is concerned.
+            binary:bin_to_list(Vsn, {0, Size - 3});
+        _ ->
+            binary:bin_to_list(Vsn, {0, Size - 1})
+    end.
+
 get_deprecated_3(Get, Config, OldOpt, NewOpt, Default, When) ->
     case Get(Config, NewOpt, Default) of
         Default ->
@@ -328,15 +481,28 @@ get_deprecated_3(Get, Config, OldOpt, NewOpt, Default, When) ->
             New
     end.
 
+get_experimental_3(Get, Config, Opt, Default) ->
+    Val = Get(Config, Opt, Default),
+    case Val of
+        Default ->
+            Default;
+        Val ->
+            ?CONSOLE("NOTICE: Using experimental option '~p'~n", [Opt]),
+            Val
+    end.
+
 %% We do the shell variable substitution ourselves on Windows and hope that the
 %% command doesn't use any other shell magic.
 patch_on_windows(Cmd, Env) ->
     case os:type() of
         {win32,nt} ->
-            "cmd /q /c "
+            Cmd1 = "cmd /q /c "
                 ++ lists:foldl(fun({Key, Value}, Acc) ->
                                        expand_env_variable(Acc, Key, Value)
-                               end, Cmd, Env);
+                               end, Cmd, Env),
+            %% Remove left-over vars
+            re:replace(Cmd1, "\\\$\\w+|\\\${\\w+}", "",
+                       [global, {return, list}]);
         _ ->
             Cmd
     end.
@@ -377,8 +543,9 @@ log_msg_and_abort(Message) ->
 
 -spec log_and_abort(string(), {integer(), string()}) -> no_return().
 log_and_abort(Command, {Rc, Output}) ->
-    ?ABORT("~s failed with error: ~w and output:~n~s~n",
-           [Command, Rc, Output]).
+    ?ABORT("sh(~s)~n"
+           "failed with return code ~w and the following output:~n"
+           "~s~n", [Command, Rc, Output]).
 
 sh_loop(Port, Fun, Acc) ->
     receive
@@ -422,23 +589,87 @@ emulate_escript_foldl(Fun, Acc, File) ->
             Error
     end.
 
-vcs_vsn_cmd(git) ->
-    %% git describe the last commit that touched CWD
-    %% required for correct versioning of apps in subdirs, such as apps/app1
-    case os:type() of
-        {win32,nt} ->
-            "FOR /F \"usebackq tokens=* delims=\" %i in "
-                "(`git log -n 1 \"--pretty=format:%h\" .`) do "
-                "@git describe --always --tags %i";
-        _ ->
-            "git describe --always --tags `git log -n 1 --pretty=format:%h .`"
-    end;
-vcs_vsn_cmd(hg)  -> "hg identify -i";
-vcs_vsn_cmd(bzr) -> "bzr revno";
-vcs_vsn_cmd(svn) -> "svnversion";
+vcs_vsn_1(Vcs, Dir) ->
+    case vcs_vsn_cmd(Vcs) of
+        {plain, VsnString} ->
+            VsnString;
+        {cmd, CmdString} ->
+            vcs_vsn_invoke(CmdString, Dir);
+        unknown ->
+            ?ABORT("vcs_vsn: Unknown vsn format: ~p\n", [Vcs]);
+        Cmd ->
+            %% If there is a valid VCS directory in the application directory,
+            %% use that version info
+            Extension = lists:concat([".", Vcs]),
+            case filelib:is_dir(filename:join(Dir, Extension)) of
+                true ->
+                    ?DEBUG("vcs_vsn: Primary vcs used for ~s\n", [Dir]),
+                    vcs_vsn_invoke(Cmd, Dir);
+                false ->
+                    %% No VCS directory found for the app. Depending on source
+                    %% tree structure, there may be one higher up, but that can
+                    %% yield unexpected results when used with deps. So, we
+                    %% fallback to searching for a priv/vsn.Vcs file.
+                    VsnFile = filename:join([Dir, "priv", "vsn" ++ Extension]),
+                    case file:read_file(VsnFile) of
+                        {ok, VsnBin} ->
+                            ?DEBUG("vcs_vsn: Read ~s from priv/vsn.~p\n",
+                                   [VsnBin, Vcs]),
+                            string:strip(binary_to_list(VsnBin), right, $\n);
+                        {error, enoent} ->
+                            ?DEBUG("vcs_vsn: Fallback to vcs for ~s\n", [Dir]),
+                            vcs_vsn_invoke(Cmd, Dir)
+                    end
+            end
+    end.
+
+vcs_vsn_cmd(git)    -> "git describe --always --tags";
+vcs_vsn_cmd(p4)     -> "echo #head";
+vcs_vsn_cmd(hg)     -> "hg identify -i";
+vcs_vsn_cmd(bzr)    -> "bzr revno";
+vcs_vsn_cmd(svn)    -> "svnversion";
+vcs_vsn_cmd(fossil) -> "fossil info";
 vcs_vsn_cmd({cmd, _Cmd}=Custom) -> Custom;
-vcs_vsn_cmd(Version) -> {unknown, Version}.
+vcs_vsn_cmd(Version) when is_list(Version) -> {plain, Version};
+vcs_vsn_cmd(_) -> unknown.
 
 vcs_vsn_invoke(Cmd, Dir) ->
     {ok, VsnString} = rebar_utils:sh(Cmd, [{cd, Dir}, {use_stdout, false}]),
     string:strip(VsnString, right, $\n).
+
+%%
+%% Filter a list of erl_opts platform_define options such that only
+%% those which match the provided architecture regex are returned.
+%%
+filter_defines([], Acc) ->
+    lists:reverse(Acc);
+filter_defines([{platform_define, ArchRegex, Key} | Rest], Acc) ->
+    case rebar_utils:is_arch(ArchRegex) of
+        true ->
+            filter_defines(Rest, [{d, Key} | Acc]);
+        false ->
+            filter_defines(Rest, Acc)
+    end;
+filter_defines([{platform_define, ArchRegex, Key, Value} | Rest], Acc) ->
+    case rebar_utils:is_arch(ArchRegex) of
+        true ->
+            filter_defines(Rest, [{d, Key, Value} | Acc]);
+        false ->
+            filter_defines(Rest, Acc)
+    end;
+filter_defines([Opt | Rest], Acc) ->
+    filter_defines(Rest, [Opt | Acc]).
+
+cleanup_code_path(OrigPath) ->
+    CurrentPath = code:get_path(),
+    AddedPaths = CurrentPath -- OrigPath,
+    %% If someone has removed paths, it's hard to get them back into
+    %% the right order, but since this is currently rare, we can just
+    %% fall back to code:set_path/1.
+    case CurrentPath -- AddedPaths of
+        OrigPath ->
+            _ = [code:del_path(Path) || Path <- AddedPaths],
+            true;
+        _ ->
+            code:set_path(OrigPath)
+    end.
